#!/usr/bin/env ruby


# #############################################################################
#  ATTENTION
# #############################################################################
#
#  This software is for guidance only. It is provided "as is".
#  It only aims at providing coding information to the customer in order for
#  them to save time. As a result no warranties, whether express, implied or
#  statutory, including, but not limited to implied warranties of
#  merchantability and fitness for a particular purpose apply to this software.
#  A·P·E Angewandte Physik und Elektronik GmbH shall not be held liable for any
#  direct, indirect or consequential damages with respect to any claims arising
#  from the content of such sourcecode and/or the use made by customers of the
#  coding information contained herein in connection with their products.
#
#  File Version: 1.0
#
#  Changelog:
#    1.0 - Initial Version
#
# #############################################################################

#-------------------------------------------------------------------------------
# REQUIRED LIBRARIES / INCLUDES
#-------------------------------------------------------------------------------
require 'socket' 
require 'timeout'	

include Socket::Constants

#-------------------------------------------------------------------------------
# APE_DEVICE Class
#-------------------------------------------------------------------------------
class APE_DEVICE

	#---------------------------------------------------------------------------
	# CLASS INSTANCE VARIABLES
	#---------------------------------------------------------------------------
	attr_reader	:dev 	#TCPSocket Object
	attr_reader	:timeout#Socket Read-Timeout (in seconds)
	attr_reader	:host	#TCP Connection Host (can be DNS-Name or IP)
	attr_reader	:port	#TCP Connection Port (1..65535)
	attr_reader	:name	#User definable device name
	
	#---------------------------------------------------------------------------
	#Get connected state
	#
	#Parameters:
	#	none
	#
	#Return:
	#	true, if connection is active
	#	false, if no active connection exists
	#
	#Throws:
	#	 none
	#---------------------------------------------------------------------------
	def connected?
		return false if !@dev
		return true
	end
	
	#---------------------------------------------------------------------------
	#Connect to specified host
	#Function is only used internally
	#
	#Parameters:
	#	none
	#
	#Return:
	#	true, if connection is active
	#	false, connection could not be established
	#
	#Throws:
	#	StandardError
	#---------------------------------------------------------------------------
	def connect
		if connected?
			raise StandardError, "[Connect] Already connected"
			return -1
		end
		if !host
			raise ArgumentError, "[Connect] Hostname or IP must not be empty"
			return -1
		end
		if !port
			raise ArgumentError, "[Connect] Port must be a positive numeric value"
			return -1
		end
		
		socket = Socket.new(AF_INET, SOCK_STREAM, 0)
		socketaddr = Socket.sockaddr_in(@port, @host)
		
		begin
			Timeout::timeout(@timeout) do
				begin
					
					socket.connect_nonblock(socketaddr)		
				rescue IO::WaitWritable		
					IO.select(nil, [socket])
					begin
						socket.connect_nonblock(socketaddr) 
					rescue Errno::EISCONN 
						@dev = socket
					end
				rescue => e
					raise e
					@dev = nil
					return -1
				end
			end
		rescue Timeout::Error
			return 1
		end
		#Wait 1 second before continuing after successful connect
		sleep 1
		
		return 0
	end

	#---------------------------------------------------------------------------
	#Function closes the socket connection and sets the global variable invaid
	#
	#Parameter:
	#	none
	#
	#Return:
	#	true, if disconnect was successful
	#	false, if an error occurred
	#
	#Throws:
	#	StandardError
	#---------------------------------------------------------------------------
	def disconnect
		if !connected?
			raise StandardError, "[Disconnect] Not connected"
			return false
		end
		
		@dev.close
		@dev = nil
		return true
	end

	#---------------------------------------------------------------------------
	#APE_Device Constructor
	#
	#Parameter:
	#	host - Hostname or IP-Address as string
	#	port - Portnumber
	#	name - (optional) Custom Name for the device
	#
	#Return:
	#	none
	#
	#Throws: 
	#	ArgumentError
	#	StandardError
	#	IOError
	#---------------------------------------------------------------------------
	def initialize host, port, name="APEDevice"

		if (host.class != String)
			raise ArgumentError, "[INIT] Hostname must be passed as string"
			return
		end
		if (host=="")
			raise ArgumentError, "[INIT] Hostname or IP must not be empty"
			return
		end
		if (port.class != Fixnum)
			raise ArgumentError, "[INIT] Port must be a positive numeric value"
			return
		end
		if (port <= 0)
			raise ArgmentError, "[INIT] Portnumber must be within 0..65535"
			return
		end
		if (name.class != String)
			raise ArgumentError, "[INIT] Devicename must be passed as string"
			return
		end

		@host = host
		@port = port
		@name = name
		@timeout = 5.0
		result = 1
		while (result > 0)
			result = connect()
			if result == 1
				puts "[Connect] Unable to connect to the specified host. Retry (Y/n)?"
				choice = gets().chomp
				if !(choice.empty? ||  choice.index(/[Yy]/))
					result = 0
				end
			end
		end
		
	end
	
	#---------------------------------------------------------------------------
	#APE_Device Destructor
	#
	#Parameter:
	#	none
	#
	#Return:
	#	none
	#
	#Throws:
	#	StandardError
	#---------------------------------------------------------------------------
	def finalize
		disconnect() if connected?
	end
	
	#---------------------------------------------------------------------------
	#Set hostname
	#
	#Parameter:
	#	h - Hostname as string
	#
	#Return:
	#	New hostname
	#
	#Throws:
	#	ArgumentError
	#	StandardError
	#---------------------------------------------------------------------------
	def set_host h
		if port.class != String
			raise ArgumentError, "[Set Host] Hostname must be passed as string"
			return
		end		
		disconnect() if connected?
		@host = h
		return @host
	end
	
	#---------------------------------------------------------------------------
	#Set portnumber
	#
	#Parameter:
	#	p - Portnumber as numeric value
	#
	#Return:
	#	New portnumber
	#
	#Throws:
	#	ArgumentError
	#	StandardError
	#---------------------------------------------------------------------------
	def set_port p
		if (port.class != Fixnum)
			raise ArgumentError, "[Set Port] Port must be a positive numeric value"
			return
		end
		if (port <= 0)
			raise ArgmentError, "[Set Port] Portnumber must be within 0..65535"
			return
		end
		disconnect() if connected?
		@port = p
		return @port
	end
	
	#---------------------------------------------------------------------------
	#Set socket read/write timeout
	#
	#Parameter:
	#	t - Timeout in seconds
	#
	#Return:
	#	New Timeout Value in seconds
	#
	#Throws:
	#	ArgumentError
	#	StandardError
	#---------------------------------------------------------------------------
	def set_timout t
		if (t.class != Fixnum)&&(t.class != Float)
			raise ArgumentError, "[Set Timeout] Timeout must be a numeric or float value"
			return
		end
		@timeout = t
		return @timeout
	end
	
	#---------------------------------------------------------------------------
	#Write a command to the device
	#
	#Parameters:
	#	command - Command String to send
	#
	#Return: 
	#	Number of bytes sent
	#
	#Throws:
	#	ArgumentError
	#	StandardError
	#---------------------------------------------------------------------------
	def send command
		len = 0
		if !connected?
			raise StandardError, "[Send] Not conncted"
			return 0
		end
		
		if command.class != String
			raise ArgumentError, "[Send] command must be passed as string"
			return 0
		end	
		begin
			Timeout::timeout(@timeout) do
				len = @dev.write(command.chomp+"\r\n")
			end
		rescue Timeout::Error 
			raise IOError, "[Send] Connection Timeout"
			return 0
		end
		
		return len
	end
	
	#---------------------------------------------------------------------------
	#Read a specified amount of data from the device
	#
	#Parameters:
	#	length - Number of characters to read
	#
	#Return: 
	#	Received data as ruby-string
	#
	#Throws:
	#	ArgumentError
	#	StandardError
	#	IOError
	#---------------------------------------------------------------------------
	def receive length
		data = ""
		if !connected?
			raise StandardError, "[Receive] Not conncted"
			return ""
		end
		
		if length.class != Fixnum
			raise ArgumentError, "[Receive] Length must be passed as numeric value"
			return ""
		end	
		begin
			Timeout::timeout(@timeout) do
				data = @dev.read(length)
			end
		rescue Timeout::Error 
			raise IOError, "[Receive] Connection Timeout"
			return ""
		end
		return data
	end
	
	#---------------------------------------------------------------------------
	#Read one line of text from the socket (\n is termination character)
	#
	#Parameters:
	#	none
	#
	#Return: 
	#	Received data as ruby-string
	#
	#Throws:
	#	StandardError
	#	IOError
	#---------------------------------------------------------------------------
	def read_line
		line = ""
		if !connected?
			raise StandardError, "[Read Line] Not conncted"
			return ""
		end
		begin
			Timeout::timeout(@timeout) do
				line = @dev.gets
			end
		rescue Timeout::Error 
			raise IOError, "[Read Line] Connection Timeout"
			return ""
		end
		return line
	end

	#---------------------------------------------------------------------------
	#Read and decode a scpi block from the socket
	#
	#Parameters:
	#	none
	#
	#Return: 
	#	No error: Data contained in the scpi block as ruby-string
	#	Error: Empty String
	#
	#Throws: 
	#	StandardError
	#	IOError
	#---------------------------------------------------------------------------
	def read_scpi

		block_data = ""	

		if !connected?
			raise StandardError, "[Read SCPI Block] Not conncted"
			return ""
		end
		#Read #character
		receive(1)
		
		header_len = receive(1)

		if  header_len[/[^\d]+/] 
			raise IOError, "[Read SCPI Block] Expected decimal string; Got: #{header_len}"
			return ""
		end

		data_len = receive(header_len.to_i(10)) 
		if  data_len[/[^\d]+/] 
			raise IOError, "[Read SCPI Block] Expected decimal string; Got: #{data_len}"
			return ""
		end

		block_data = receive(data_len.to_i(10))
		return block_data
	end
	
	#---------------------------------------------------------------------------
	#Request data from the device by sending a command
	#
	#Parameters:
	#	command 	INPUT 			Command String to send
	#	read_block	INPUT(optional) If set to true, function will attempt to
	#  								read and unpack SCPI block data
	#Return: 
	#	Query successful: Received data as ruby-string
	#	Error: Empty string
	#
	#Throws:
	#	IOError
	#	StandardError
	#---------------------------------------------------------------------------
	def query command, read_block = false

		data = ""
		
		if !connected?
			raise StandardError, "[Query] Not conncted"
			return ""
		end
		begin
			Timeout::timeout(@timeout) do
				send(command)
				
				if read_block
					data = read_scpi
				else
					data = @dev.gets()
				end
			end
		rescue Timeout::Error 
			raise IOError, "[Query] Connection Timeout"
			return ""
		end
		
		return data
	end
	
	#---------------------------------------------------------------------------
	#Request SCPI *IDN?
	#
	#Parameters:
	#	none
	#
	#Return: 
	#	Received IDN-String
	#
	#Throws:
	#	IOError
	#	StandardError
	#---------------------------------------------------------------------------
	def idn? 

		idn = ""
		
		if !connected?
			raise StandardError, "[IDN?] Not conncted"
			return ""
		end
		
		begin
			Timeout::timeout(@timeout) do
				send("*IDN?")
				idn = read_line()
			end
		rescue Timeout::Error 
			raise IOError, "[IDN?] Connection Timeout"
			return ""
		end
		
		return idn
	end

	#---------------------------------------------------------------------------
	#Request SCPI *STB?
	#
	#Parameters:
	#	none
	#
	#Return: 
	#	Received Status Register Value
	#
	#Throws:
	#	IOError
	#	StandardError
	#---------------------------------------------------------------------------
	def status?

		status = ""
		
		if !connected?
			raise StandardError, "[STATUS?] Not conncted"
			return ""
		end
		
		begin
			Timeout::timeout(@timeout) do
				send("*STB?")
				status = read_line()
			end
		rescue Timeout::Error 
			raise IOError, "[STATUS?] Connection Timeout"
			return ""
		end
		
		return status
	end
	
	#---------------------------------------------------------------------------
	#Request SCPI *OPER?
	#
	#Parameters:
	#	none
	#
	#Return: 
	#	Received Operation Status Register Value
	#
	#Throws:
	#	IOError
	#	StandardError
	#---------------------------------------------------------------------------
	def oper?

		status = ""
		
		if !connected?
			raise StandardError, "[OPER?] Not conncted"
			return ""
		end
		
		begin
			Timeout::timeout(@timeout) do
				send("*OPER?")
				status = read_line()
			end
		rescue Timeout::Error 
			raise IOError, "[OPER?] Connection Timeout"
			return ""
		end
		
		return status
	end
end #END CLASS ape_device